import bz2
import codecs
import json
import re
import unicodedata
import os.path

import pytest
from requests.structures import CaseInsensitiveDict

from guessenc import infer_encoding, Pair, Source

DATA_DIR = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "data")
assert os.path.exists(DATA_DIR), "Data directory not found"


def slugify(value: str) -> str:
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


data = [
    (
        "http://cen.ce.cn/more/201909/23/t20190923_33202170.shtml",
        (Source.META_CHARSET, codecs.lookup("gb2312").name),
        (Source.COULD_NOT_DETECT, None),
        (Source.META_CHARSET, codecs.lookup("gb2312").name),
    ),
    (
        "http://www.arirang.com/news/News_View.asp?nseq=244581",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://www.nytimes.com/2019/09/23/us/politics/trump-un-biden-ukraine.html",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://www.currenttime.tv/a/semsorok-irrational/30180436.html",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "http://www.xinhuanet.com/politics/70zn/fbh2/index.htm",
        (Source.META_HTTP_EQUIV, codecs.lookup("utf-8").name),
        (Source.COULD_NOT_DETECT, None),
        (Source.META_HTTP_EQUIV, codecs.lookup("utf-8").name),
    ),
    (
        "http://www.sanaanews.net/news-66176.htm",
        (Source.META_HTTP_EQUIV, codecs.lookup("cp1256").name),
        (Source.COULD_NOT_DETECT, None),
        (Source.META_HTTP_EQUIV, codecs.lookup("cp1256").name),
    ),
    (
        "http://alfajertv.com/news/4037470.html",
        (Source.META_HTTP_EQUIV, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://www.presstv.com/Detail/2019/09/23/606960/Rouhani-newyork-hope-initiative-hormuz-peace",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://www.dap-news.com/archives/70763 ",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "http://www.fatehwatan.ps/page-183525.html",
        (Source.META_HTTP_EQUIV, codecs.lookup("windows-1256").name),
        (Source.COULD_NOT_DETECT, None),
        (Source.META_HTTP_EQUIV, codecs.lookup("windows-1256").name),
    ),
    (
        "https://klse.i3investor.com/blogs/kianweiaritcles/211663.jsp",
        (Source.META_HTTP_EQUIV, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://www.theguardian.com/society/2018/dec/25/"
        "doubletree-steps-in-after-hotel-cancels-booking-for-homeless-people",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "http://koreabizwire.com/bts-goes-beyond-barriers-through-web-based-content/139261",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "http://www.ashburtononline.co.nz/site/news/news-headlines/"
        "us-china-trade-war-officials-to-resume-talks-before-g20.html",
        (Source.META_HTTP_EQUIV, codecs.lookup("iso-8859-1").name),
        (Source.COULD_NOT_DETECT, None),
        (Source.META_HTTP_EQUIV, codecs.lookup("iso-8859-1").name),
    ),
    (
        "https://www.aljazeera.net/news/politics/2019/9/23/"
        "%D9%87%D8%AC%D9%88%D9%85-%D8%A3%D8%B1%D8%A7%D9%85%D9%83%D9%88-"
        "%D8%A5%D9%8A%D8%B1%D8%A7%D9%86-%D8%A8%D9%8A%D8%A7%D9%86-%D8%AB%D9%84%D8%A7%D8%AB%D9%8A",
        (Source.META_CHARSET, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
    (
        "https://api.gdeltproject.org/api/v2/doc/doc?mode=artList&format=JSON&maxrecords=10&query=Litvinenko",
        (Source.COULD_NOT_DETECT, None),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
        (Source.CHARSET_HEADER, codecs.lookup("utf-8").name),
    ),
]


@pytest.mark.parametrize("url,with_content_only,with_headers_only,with_content_and_headers", data)
def test_infer_encoding(url: str, with_content_only: Pair, with_headers_only: Pair, with_content_and_headers: Pair):
    slug = slugify(url)
    html_fp = os.path.join(DATA_DIR, "html", slug + ".bz2")
    headers_fp = os.path.join(DATA_DIR, "headers", slug)

    content = bz2.decompress(open(html_fp, "rb").read())
    with open(headers_fp) as f:
        headers = CaseInsensitiveDict(json.load(f))

    assert infer_encoding(content=content, headers=None) == with_content_only
    assert infer_encoding(content=None, headers=headers) == with_headers_only
    assert infer_encoding(content=content, headers=headers) == with_content_and_headers



#!/usr/bin/env python3

import argparse
import json
import urllib.parse

import requests

from tests.test_guessenc import slugify

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:69.0) Gecko/20100101 Firefox/69.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Connection": "keep-alive",
    "Accept-Encoding": "gzip, deflate",
}


def download(url):
    slug = slugify(url)
    headers = HEADERS.copy()
    headers["Host"] = urllib.parse.urlparse(url).netloc
    resp = requests.get(url, headers=headers)
    resp.raise_for_status()
    with open(f"data/html/{slug}", "wb") as f:
        f.write(resp.content)
    with open(f"data/headers/{slug}", "w") as f:
        json.dump(dict(resp.headers), f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("url", nargs="+", help="URL string")
    args = parser.parse_args()
    for u in args.url:
        print("Downloading", u)
        download(u)
        print("...Done")



"""Guess encoding from response headers and/or page content."""

__all__ = ("infer_encoding",)
__version__ = "0.3"

import codecs
import enum
from typing import Mapping, Optional, Tuple

import chardet
import lxml.html


class Source(enum.Enum):
    """Indicates where our detected encoding came from."""

    CHARSET_HEADER = 0
    META_CHARSET = 1
    META_HTTP_EQUIV = 2
    CHARDET = 3
    COULD_NOT_DETECT = 4


Pair = Tuple[Source, Optional[str]]


def infer_encoding(content: Optional[bytes] = None, headers: Optional[Mapping[str, str]] = None) -> Pair:
    """Infer encoding from response headers and/or page content.

    :param content: The page HTML (optional), such as ``response.content``.
    :type content: bytes

    :param headers: The response headers, such as ``response.headers``.
        This should be a data structure supporting a case-insensitive
        lookup, such as ``requests.structures.CaseInsensitiveDict``
        or ``multidict.CIMultiDict``.
    :type headers: Mapping[str, str]

    :return: Pair of (_source_, _encoding_).  The source is a :class:`Source`
        enum member that indicates where in the content or header the encoding
        was detected from.  The encoding is a ``str`` representing the
        formal codec name, or ``None`` if no encoding can be detected.
    :rtype: Tuple[Source, Optional[str]]

    Example:

    >>> import requests
    >>> resp = requests.get("http://www.fatehwatan.ps/page-183525.html")
    >>> resp.raise_for_status()
    >>> infer_encoding(resp.content, resp.headers)
    (<Source.META_HTTP_EQUIV: 2>, 'cp1256')

    This tells us that the detected encoding is cp1256, and that it
    was retrieved from a <meta> HTML tag with
    ``http-equiv='Content-Type'``.
    """

    headers = headers or {}
    src, res = infer_from_headers(headers)
    if not res:
        if not content:
            return Source.COULD_NOT_DETECT, None
        src, res = infer_from_content(content)
    if res:
        return src, codecs.lookup(res).name
    return Source.COULD_NOT_DETECT, None


def infer_from_content(content: bytes) -> Pair:
    # - <meta charset="UTF-8" />
    # - <meta charset="utf-8" />
    # - <meta charset="utf-8">
    # - <meta charset="gb2312">
    doc = lxml.html.fromstring(content)
    el = doc.find(".//meta[@charset]")
    if el is not None:
        val = el.get("charset")
        if val is not None:
            return Source.META_CHARSET, val

    # - <meta content='text/html; charset=UTF-8' http-equiv='Content-Type'/>
    # - <meta content='text/html; charset=windows-1256' http-equiv='Content-Type' />
    # - <meta http-equiv="Content-Type" content="text/html; charset=cp1256">
    # - <meta http-equiv="Content-Type" content="text/html; charset='LATIN-1'">
    # - <meta http-equiv="Content-Type" content="text/html; charset=gb2312" />
    # N.B.: HTML attribute values *are* case-sensitive.  We also cannot
    # use bool(el)
    for path in (".//meta[@http-equiv='Content-Type']", ".//meta[@http-equiv='content-type']"):
        el = doc.find(path)
        found = el is not None
        if found:
            break
    if found:
        val = el.get("content")
        if val is not None:
            return Source.META_HTTP_EQUIV, parse_mimetype_charset_param(val)

    try:
        return Source.CHARDET, chardet.detect(val)["encoding"]
    except Exception:
        return Source.COULD_NOT_DETECT, None


def infer_from_headers(headers: Mapping[str, str]) -> Pair:
    mtype = headers.get("Content-Type")
    if not mtype:
        return Source.COULD_NOT_DETECT, None
    parsed_mtype = parse_mimetype_charset_param(mtype)
    if parsed_mtype:
        return Source.CHARSET_HEADER, parsed_mtype
    return Source.COULD_NOT_DETECT, None


def parse_mimetype_charset_param(mimetype: str) -> Optional[str]:
    parts = mimetype.split(";")
    for item in parts[1:]:
        if not item:
            continue
        try:
            key, value = item.split("=", 1)
        except ValueError:
            pass
        if key.casefold().strip() == "charset":
            return value.strip(" \"'")
    if parts[0] == "application/json":
        # RFC 7159: default JSON encoding is UTF-8
        return "utf-8"
    return None



import os
from setuptools import setup

from guessenc import __version__

# PyPI:
# $ python setup.py sdist bdist_wheel
# $ twine upload dist/guessenc-x.y

here = os.path.abspath(os.path.dirname(__file__))

if __name__ == "__main__":
    setup(
        name="guessenc",
        version=__version__,
        author="Brad Solomon",
        author_email="brad.solomon.1124@gmail.com",
        description="Infer HTML encoding from response headers & content",
        license="MIT",
        keywords="encoding http html chardet detection",
        url="https://github.com/bsolomon1124/guessenc",
        long_description=open(os.path.join(here, "README.md"), encoding="utf-8").read(),
        long_description_content_type="text/markdown",
        install_requires=["lxml", "chardet"],
        packages=["guessenc"],
        tests_require=["pytest"],
        python_requires=">=3.5",
        classifiers=[
            "Intended Audience :: Developers",
            "Development Status :: 3 - Alpha",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: Implementation :: CPython",
        ],
    )
